#!/usr/bin/python2.3

"""PXE Imager
By Evan Jones
http://evanjones.ca/software/pxeimager.html

This script grabs a list of images, presents a menu, then installs the image.
It requires "atftp" as a client to fetch files via tftp.


INIT SCRIPT REQUIREMENTS:
- Mount the root at /mnt

IMAGE REQUIREMENTS:
- Tar file that gets extracted to /mnt

POST SCRIPT REQUIREMENTS:
- Make the device bootable

Imager will unmount and reboot after executing the post script
""" 

import sys
import os
import os.path
import traceback

# Paths and commands
tempDir = "/tmp"
mountDir = "/mnt"
tarCommand = "tar xpf - --numeric-owner &> /dev/null"
umount = "/bin/umount"
reboot = "/sbin/reboot"
sleep = "/bin/sleep"
init = "/sbin/init"
shell = "/bin/bash"
systemboot = "/etc/init.d/rcS"
dhcpClientLeases = "/var/run/dhclient.eth0.leases"

tftpCommand = 'atftp %s --get --remote-file %s --local-file /proc/self/fd/1 --option "blksize 1428"'

# Hard coded names that must exist on the tftp server
imageList = "images"
initScriptName = "init"
imageName = "image.tar"
postScriptName = "post"

def execShell():
	'''Replaces this Python process with a shell.'''
	os.execl( shell, shell )

def tftp( server, file ):
	'''Gets file from server using TFTP. It returns a file object that will
	get data from the server. It is implemented as a pipe from an atftp
	process.'''
	return os.popen( tftpCommand % ( server, file ) )

def URLToExecutable( tftpServer, path, exe ):
	'''Gets path from tftpServer and writes it to exe. Finally, it makes
	exe executable.'''
	
	output = open( exe, "w" )
	
	input = tftp( tftpServer, path )
	
	output.write( input.read() )
	
	input.close()
	output.close()
	
	# Make it executable
	os.chmod( exe, 0755 )

# Put this whole thing in a try/except block so we drop to the shell if an
# exception occurs
try:
	

	# Determine the address of the DHCP server
	f = None
	try:
		f = open( dhcpClientLeases )
	except IOError:
		print "FATAL ERROR: Could not open %s" % ( dhcpClientLeases )
		raise

	serverIP = None
	for line in f:
		# Trim whitespace
		line = line.strip()
		
		# Trim "option" off the beginning
		option = "option"
		if line.startswith( option ):
			line = line[ len(option): ].strip()
			
			server = "dhcp-server-identifier"
			if line.startswith( server ):
				line = line[ len(server): ].strip()
			
				if line.endswith( ";" ):
					line = line[:-1].strip()
						
					# What is left should be the IP address
					serverIP = line
					break
	f.close()

	if serverIP == None:
		print "FATAL ERROR: Could not locate DHCP server"
		execShell()

	images = []
	data = tftp( serverIP, imageList )
	for line in data:
		# Strip whitespace from both ends of the image
		line = line.strip()
		
		# Skip blank lines and comments
		if line == "" or line[0] == '#':
			continue
		
		# Split the line on the first whitespace
		line = line.split( None, 1 )
		if len( line ) != 2 or line[0] == "" or line[1] == "":
			print "FATAL ERROR: Image file not in the correct format"
			execShell()
		images.append( line )
	data.close()

	selectedImage = 0

	# Select an image to install
	if len( images ) == 0:
		print "FATAL ERROR: No images!"
		execShell()
	elif len( images ) > 1:
		# Print the list of available images
		selectedImage = -1
		while selectedImage == -1:
			for i in range( len( images ) ):
				print "%d: %s" % ( i, images[i][1] )
			print "s: Execute Shell"
			input = raw_input( "Enter your selection: " )
			
			if input == "s":
				execShell()
			
			try:
				selectedImage = int( input )
				if not (0 <= selectedImage and selectedImage < len( images )):
					selectedImage = -1
			except ValueError:
				pass
			
			print
			if selectedImage == -1:
				print "Invalid selection. Try again."

	print "Installing %s ..." % ( images[selectedImage][1] )

	# Create temporary directory
	realTempDir = os.path.join( tempDir, str( os.getpid() ) )
	os.mkdir( realTempDir )

	# Get the URLs
	initScriptPath = os.path.join( realTempDir, initScriptName )
	URLToExecutable( serverIP, os.path.join( images[selectedImage][0], initScriptName ), initScriptPath )

	postScriptPath = os.path.join( realTempDir, postScriptName )
	URLToExecutable( serverIP, os.path.join( images[selectedImage][0], postScriptName ), postScriptPath )

	# Execute the initscript
	code = os.spawnl( os.P_WAIT, initScriptPath, initScriptPath )
	if code != 0:
		print "FATAL ERROR: Init Script returned code %d" % ( code )
		execShell()

	# Extract the image
	os.chdir( mountDir )
	
	# Create a pipe from aftp to tar
	tftpPipeCommand = tftpCommand % ( serverIP, os.path.join( images[selectedImage][0], imageName ) )
	code = os.system( "%s | %s" % ( tftpPipeCommand, tarCommand ) )
	if code != 0:
		print "FATAL ERROR: tftp/tar returned code %d" % ( code )
		execShell()
	os.chdir( "/" )

	# Execute the postscript
	code = os.spawnl( os.P_WAIT, postScriptPath, postScriptPath )
	if code != 0:
		print "FATAL ERROR: Post script returned code %d" % ( code )
		execShell()
		
	code = os.spawnl( os.P_WAIT, umount, umount, mountDir )
	if code != 0:
		print "FATAL ERROR: Unmounting %s failed code %d" % ( mountDir, code )
		execShell()

	# We are all done! Reboot!
	print "Installation Successful. Rebooting in 3 seconds ..."
	
	# Delay for a moment
	code = os.spawnl( os.P_WAIT, sleep, sleep, "3" )
	assert code == 0
	
	os.execl( reboot, reboot, "-f" )

except:
	print traceback.print_exc()
	execShell()
